@@ -1,3 +1,4 @@
+#[cfg(test)] mod proptest;
 #[cfg(test)] mod test;
 
 pub trait Item {
@@ -26,6 +27,7 @@ impl<I: crate::Item, T> Item for Tagged<I, T> {
 	}
 }
 
+#[derive(Debug)]
 pub struct Bucket<Period> {
 	period: Period,
 	count: usize,
@@ -40,6 +42,7 @@ impl<Period> Bucket<Period> {
 	}
 }
 
+#[derive(Debug)]
 pub struct Config<Period> {
 	buckets: Vec<Bucket<Period>>,
 }
@@ -77,7 +80,9 @@ impl<Period> Config<Period> {
 #[derive(Debug)]
 struct BucketState {
 	remaining_slots: usize,
+	/// The first interesting item, relative to the previous bucket's anchor.
 	anchor: usize,
+	/// The last owned item, relative to the previous bucket's last.
 	last: usize,
 }
 
@@ -89,6 +94,7 @@ struct Iter<'a, Item: crate::Item, I: Iterator<Item=Item>> {
 
 	next_bucket: usize,
 	anchor_ref: usize,
+	last_ref: usize,
 }
 
 impl<'a, Item: crate::Item, I: Iterator<Item=Item>> Iter<'a, Item, I> {
@@ -107,25 +113,21 @@ impl<'a, Item: crate::Item, I: Iterator<Item=Item>> Iter<'a, Item, I> {
 
 	fn disown(&mut self, bid: usize) -> Option<Item> {
 		let bucket_state = &mut self.buckets[bid];
-		let last = bucket_state.last;
+		let last = self.last_ref;
 		if let Some(bucket) = self.config.buckets.get(bid+1) {
 			if last > 0 && !self.keep[last].matches(&self.keep[last-1], &bucket.period) {
-				for b in &mut self.buckets[..bid] {
-					b.last -= 1;
-				}
+				bucket_state.last -= 1;
 				self.adjust_anchors(last, bid);
 				Some(self.keep.remove(last))
 			} else {
-				bucket_state.last += 1;
+				bucket_state.last -= 1;
+				self.buckets[bid+1].last += 1;
+				self.last_ref += 1; // To account for the fact that we just incremented the rel_last for the next bucket.
 				self.next()
 			}
 		} else {
 			debug_assert_eq!(last, 0);
-			for b in 0..bid {
-				if self.buckets[b].last > 0 {
-					self.buckets[b].last -= 1;
-				}
-			}
+			bucket_state.last = bucket_state.last.saturating_sub(1);
 			Some(self.keep.remove(last))
 		}
 	}
@@ -143,6 +145,7 @@ impl<'a, Item: crate::Item, I: Iterator<Item=Item>> Iterator for Iter<'a, Item,
 
 			self.next_bucket += 1;
 			self.anchor_ref -= bucket_state.anchor;
+			self.last_ref -= bucket_state.last;
 
 			let anchor = &self.keep[self.anchor_ref];
 			if !self.keep.last().unwrap().matches(anchor, &bucket_config.period) {
@@ -167,6 +170,10 @@ impl<'a, Item: crate::Item, I: Iterator<Item=Item>> Iterator for Iter<'a, Item,
 
 		let e = self.candidates.next()?;
 
+		if self.config.buckets.is_empty() {
+			return Some(e)
+		}
+
 		if self.keep.is_empty() {
 			self.keep.push(e);
 			return self.next()
@@ -181,8 +188,13 @@ impl<'a, Item: crate::Item, I: Iterator<Item=Item>> Iterator for Iter<'a, Item,
 
 		self.next_bucket = 1;
 		self.anchor_ref = self.keep.len() - 2;
+		self.last_ref = self.keep.len()
+			.saturating_sub(first_bucket.count)
+			.saturating_sub(1); // The one that we just added.
 
 		if self.keep.len() > first_bucket.count {
+			self.buckets[0].last = 1; // Fake, but avoid integer underflow.
+			// self.last_ref -= 1;
 			self.disown(0)
 		} else {
 			self.next()
@@ -209,5 +221,6 @@ pub fn prune<'a, Item: crate::Item + 'a>(
 
 		next_bucket: config.buckets.len(),
 		anchor_ref: 0,
+		last_ref: 0,
 	}
 }
